#include "cmdline_maxlink-load.h"

#include <cstdlib>
#include <stdint.h>
#include <iostream>
#include <algorithm>
#include <set>
#include <boost/foreach.hpp>
#include <fstream>

#include <sys/stat.h>

#include "fbs_database.h"
#include "edge_database.h"
#include "num2string_gene_database.h"
#include "string2num_gene_database.h"
#include "maxlink_xml_parser.h"
#include "maxlink_text_parser.h"
#include "maxlink_types.h"

#include "shortest_path_database.h"
#include "gene_shortest_path_count_database.h"
#include "genepair_shortest_path_database.h"
#include "shortest_paths.h"
#include "common_funcs.h"

int main(int argc, char *argv[]) {
  EdgeDatabase edge_database;
  FbsDatabase fbs_database;
  Num2StringGeneDatabase num2string_gene_database;
  String2NumGeneDatabase string2num_gene_database;
  struct gengetopt_args_info args_info;
  if(cmdline_parser(argc, argv, &args_info) != 0) {
    exit(EXIT_FAILURE);
  }

  char *tokyo_cabinet_dir = get_tokyo_cabinet_dir(args_info.tokyo_cabinet_dir_given, args_info.tokyo_cabinet_dir_arg);
  assert(tokyo_cabinet_dir);

  std::string fbs_cutoff_file = std::string(tokyo_cabinet_dir) + std::string("/fbs_cutoff");
  std::ifstream test_file_existence;
  test_file_existence.open(fbs_cutoff_file.c_str()); 
  if (test_file_existence.good()) {
    std::cerr << "error: The tokyo cabinet dir is not empty" << std::endl;
    exit(EXIT_FAILURE);
  }
  std::ofstream cutoff_stream;
  cutoff_stream.open(fbs_cutoff_file.c_str()); 
  if (!cutoff_stream.good()) {
    std::cerr << "error: Could not create a file in the tokyo cabinet dir" << std::endl;
    exit(EXIT_FAILURE);
  }
  cutoff_stream <<  args_info.fbs_cutoff_arg << std::endl;

  ShortestPathDatabase shortest_path_database;
  GeneShortestPathCountDatabase gene_shortest_path_count_database;
  GenePairShortestPathDatabase endgene_shortest_path_database;
  shortest_path_database.create(tokyo_cabinet_dir);
  gene_shortest_path_count_database.create(tokyo_cabinet_dir);
  endgene_shortest_path_database.create(tokyo_cabinet_dir);
  num2string_gene_database.create(tokyo_cabinet_dir);
  string2num_gene_database.create(tokyo_cabinet_dir);
  edge_database.create(tokyo_cabinet_dir);
  fbs_database.create(tokyo_cabinet_dir);
  maxlink_types::GeneId loaded_genes_count = 0;
  switch (args_info.protein_network_file_format_arg) {
    case protein_network_file_format_arg_tsv: {
      MaxlinkTextParser text_parser;
      text_parser.loadKeyValueDatabaseFromText(&num2string_gene_database, &string2num_gene_database, &edge_database, &fbs_database, args_info.protein_network_filename_arg, &loaded_genes_count, args_info.fbs_cutoff_arg);
      break;
    }
    case protein_network_file_format_arg_xml: {
      MaxlinkXmlParser maxlink_xml_parser;
      maxlink_xml_parser.loadKeyValueDatabaseFromXml(args_info.debug_flag, &num2string_gene_database, &string2num_gene_database, &edge_database, &fbs_database, args_info.protein_network_filename_arg, &loaded_genes_count, args_info.fbs_cutoff_arg);
      break;
    }
    default: assert(false); /* indicate programming error */ 
  }
  fbs_database.calculatePositions();
  if (args_info.verbose_flag) {  
    std::cerr << "gene count = " << num2string_gene_database.num_entries() << std::endl;
    std::cerr << "edge count over fbs cutoff (" << args_info.fbs_cutoff_arg << ") = " << fbs_database.num_edges_over_cutoff(args_info.fbs_cutoff_arg) << std::endl;
  }
  if (args_info.debug_flag) {  
    std::cerr << "debug: fbs_database.num_edges_over_cutoff(args_info.fbs_cutoff_arg) = " << fbs_database.num_edges_over_cutoff(args_info.fbs_cutoff_arg) << std::endl;
  }
  if (! args_info.exclude_maxlink_version2_flag) {
    storeShortestPaths(args_info.debug_flag, &num2string_gene_database, edge_database, &gene_shortest_path_count_database, &endgene_shortest_path_database, &shortest_path_database, loaded_genes_count, 8.0);
  }
  return EXIT_SUCCESS;
}
