#include "cmdline_maxlink-query2.h"

#include <cstdlib>
#include <stdint.h>
#include <iostream>
#include <algorithm>
#include <set>
#include <boost/foreach.hpp>
#include <boost/test/utils/foreach.hpp>
#include <boost/math/distributions/hypergeometric.hpp>
#include <boost/math/policies/policy.hpp>

#include "fbs_database.h"
#include "edge_database.h"
#include "num2string_gene_database.h"
#include "string2num_gene_database.h"
#include "shortest_path_database.h"
#include "gene_shortest_path_count_database.h"
#include "genepair_shortest_path_database.h"

#include "config.h"
#include "maxlink_xml_parser.h"
#include "maxlink_text_parser.h"
#include "maxlink_gene_list_xml_parser.h"
#include "maxlink_types.h"
#include "common_funcs.h"



static bool sortHypergeometric(maxlink_types::GeneIdHyperGeometricResult u, maxlink_types::GeneIdHyperGeometricResult v)
{
  return u.hypergeometric_probability < v.hypergeometric_probability;
}

std::vector<maxlink_types::GeneId> geneListFromUserInput(const struct gengetopt_args_info &args_info, String2NumGeneDatabase *string2num_gene_database) {
  std::vector<maxlink_types::GeneId> gene_list;
  switch (args_info.gene_list_file_format_arg) {
    case gene_list_file_format_arg_xml: {
      MaxlinkGeneListXmlParser maxlink_gene_list_xml_parser;
      return maxlink_gene_list_xml_parser.loadGeneVectorFromXml(args_info.gene_list_filename_arg, string2num_gene_database);
      break;
    }
    case gene_list_file_format_arg_text: {
      std::string line;
      std::ifstream file(args_info.gene_list_filename_arg);
      maxlink_types::GeneId gene;
      while(std::getline(file, line)) {
        if (!string2num_gene_database->translateString2Num(line, &gene)) {
          std::cerr << "error: could not translate gene id string \"" << line << "\" to gene id number" << std::endl;     
          // exit(EXIT_FAILURE);
        } else {
          gene_list.push_back(gene);
        }
      }
      break;
    }
    default: assert(false); /* indicate programming error */ 
  }
  return gene_list;
}

std::vector<maxlink_types::GeneIdHyperGeometricResult> calculateMaxlinkVersion2(struct gengetopt_args_info &args_info, EdgeDatabase *edge_database, ShortestPathDatabase *shortest_path_database, GeneShortestPathCountDatabase *gene_shortest_path_count_database, GenePairShortestPathDatabase *endgene_shortest_path_database, Num2StringGeneDatabase *num2string_gene_database, std::vector<maxlink_types::GeneId> *q_vector, unsigned int *n1_ptr) {
  unsigned int n1 = 0;
  unsigned int n2 = shortest_path_database->num_entries();
  if (shortest_path_database->num_entries() == 0) {
    std::cerr << "The shortest path database was empty. Most probably the maxlink-load was run with the flag --exclude-maxlink-version2" << std::endl;
    exit(EXIT_FAILURE);
  }
  if (args_info.debug_flag) {
    std::cerr << "debug: shortest_path_database->num_entries() =" << shortest_path_database->num_entries() << std::endl;
  }
  std::sort(q_vector->begin(), q_vector->end());  
  std::set<maxlink_types::GeneId> q_set;
  std::copy(q_vector->begin(), q_vector->end(), std::inserter(q_set, q_set.begin()));
  std::vector<maxlink_types::ShortestPathId> shortest_paths_between_q_genes;
  uint64_t num_q_edges = 0;
  BOOST_FOREACH(const maxlink_types::GeneId q_iter1, *q_vector) {
    BOOST_FOREACH(const maxlink_types::GeneId q_iter2, *q_vector) {
      maxlink_types::GenePair end_genes;
      if (q_iter1 > q_iter2) {
        end_genes.first = q_iter1;
        end_genes.second = q_iter2;
        maxlink_types::ShortestPathId endgene_shortest_path_id;
        bool ret = endgene_shortest_path_database->shortestPath(end_genes, &endgene_shortest_path_id);
        if (ret) {          
          ++n1;
          shortest_paths_between_q_genes.push_back(endgene_shortest_path_id);
        } 
      }
    }
  }
  *n1_ptr = n1;
  if (args_info.debug_flag) {
    std::cerr << "debug: shortest_paths_between_q_genes.size() =" << shortest_paths_between_q_genes.size() << std::endl;
  }
  typedef std::map<maxlink_types::GeneId, int> ShortestPathsMembershipsMapType;
  ShortestPathsMembershipsMapType shortest_paths_memberships_map;

  BOOST_FOREACH(const maxlink_types::ShortestPathId path_id, shortest_paths_between_q_genes) {
    maxlink_types::ShortestPath path = shortest_path_database->shortestPath(path_id);
    if (args_info.debug_flag) {
      std::cerr << "debug: investigating a shortest path from shortest_paths_between_q_genes" << std::endl;
    }
    BOOST_FOREACH(const maxlink_types::GeneId gene_id, path) {
      if (args_info.debug_flag) {
        std::cerr << "debug: member gene = " << num2string_gene_database->translateNum2String(gene_id) << std::endl;
      }
          std::set<maxlink_types::GeneId>::iterator it = q_set.find(gene_id);
          if (it==q_set.end()) {
            // See http://stackoverflow.com/questions/97050/stdmap-insert-or-stdmap-find
                ShortestPathsMembershipsMapType::iterator map_iter;
                map_iter = shortest_paths_memberships_map.lower_bound(gene_id);
                if(map_iter != shortest_paths_memberships_map.end() && !(shortest_paths_memberships_map.key_comp()(gene_id, map_iter->first))) {
                  // gene_id has already been inserted to the map so we just increment the counter by one.
                  map_iter->second =  map_iter->second + 1;
                } else {                 
                  shortest_paths_memberships_map.insert(map_iter, ShortestPathsMembershipsMapType::value_type(gene_id, 1)); 
                }
          }
    }
  }
  if (args_info.debug_flag) {
    std::cerr << "debug: shortest_paths_memberships_map =" <<  shortest_paths_memberships_map.size() << std::endl;
  }
  std::vector<maxlink_types::GeneIdHyperGeometricResult> result_vector;
  BOOST_FOREACH(ShortestPathsMembershipsMapType::value_type &it, shortest_paths_memberships_map) {
    maxlink_types::GeneIdHyperGeometricResult result;
    result.gene_id = it.first;
    result.k = it.second;
    result.t = gene_shortest_path_count_database->shortestPathCount(it.first);

    // Read more about gsl_cdf_hypergeometric_Q() here:
    // http://www.gnu.org/software/gsl/manual/html_node/The-Hypergeometric-Distribution.html

    // The GNU Scientific Library functions were replaced with Boost functions. For more info see:
    // http://stackoverflow.com/questions/13274665/translate-gsl-ran-hypergeometric-pdfk-n1-n2-t-to-boostmathhypergeometri/13274706#13274706
    // result.hypergeometric_probability =  gsl_ran_hypergeometric_pdf(result.k, n1, n2, result.t) +  gsl_cdf_hypergeometric_Q(result.k, n1, n2, result.t);

    boost::math::hypergeometric_distribution<double> hg_dist(result.t, n1, n1 + n2);
    result.hypergeometric_probability = boost::math::pdf<double>(hg_dist, result.k) + (1 - boost::math::cdf<double>(hg_dist, result.k));
    if (args_info.probability_cutoff_arg >= result.hypergeometric_probability) {
      result_vector.push_back(result);
    }
  }
  std::sort(result_vector.begin(), result_vector.end(), sortHypergeometric);
  return result_vector;
}

int main(int argc, char *argv[]) {
  EdgeDatabase edge_database;
  Num2StringGeneDatabase num2string_gene_database;
  String2NumGeneDatabase string2num_gene_database;
  ShortestPathDatabase shortest_path_database;
  GeneShortestPathCountDatabase gene_shortest_path_count_database;
  GenePairShortestPathDatabase endgene_shortest_path_database;
  struct gengetopt_args_info args_info;
  if(cmdline_parser(argc, argv, &args_info) != 0) {
    exit(EXIT_FAILURE);
  }
  char *tokyo_cabinet_dir = get_tokyo_cabinet_dir(args_info.tokyo_cabinet_dir_given, args_info.tokyo_cabinet_dir_arg);
  assert(tokyo_cabinet_dir);
  unsigned int n1 = 0;
  num2string_gene_database.openReadOnly(tokyo_cabinet_dir);
  string2num_gene_database.openReadOnly(tokyo_cabinet_dir);
  edge_database.openReadOnly(tokyo_cabinet_dir);
  // fbs_database.openReadOnly(tokyo_cabinet_dir);
  shortest_path_database.openReadOnly(tokyo_cabinet_dir);
  gene_shortest_path_count_database.openReadOnly(tokyo_cabinet_dir);
  endgene_shortest_path_database.openReadOnly(tokyo_cabinet_dir);
  std::vector<maxlink_types::GeneId> q_vector = geneListFromUserInput(args_info, &string2num_gene_database);

  std::vector<maxlink_types::GeneIdHyperGeometricResult> maxlink_result = calculateMaxlinkVersion2(args_info, &edge_database, &shortest_path_database, &gene_shortest_path_count_database, &endgene_shortest_path_database, &num2string_gene_database, &q_vector, &n1);

  switch (args_info.output_format_arg) {
    case output_format_arg_geneId: {
      BOOST_FOREACH(maxlink_types::GeneIdHyperGeometricResult it, maxlink_result) {
        // The Coogle C++ Style Guide recommends fprintf instead of iostream     
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=Streams#Streams
        // But if we use printf we also need to take care of 64-bit portability:
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=64-bit_Portability#64-bit_Portability
        // 
        // Anyway just for now we use iostream
        std::cout << num2string_gene_database.translateNum2String(it.gene_id) << std::endl;      
      }
      break;   
    }
    case output_format_arg_fullResultTsv: {
      std::cout << "gene_id\tprob\tk\tn1\tt\tn2" << std::endl;   
      BOOST_FOREACH(maxlink_types::GeneIdHyperGeometricResult it, maxlink_result) {
        std::cout << num2string_gene_database.translateNum2String(it.gene_id) << "\t" 
           << it.hypergeometric_probability << "\t"   << it.k << "\t" << n1 << "\t" 
           << it.t << "\t" << shortest_path_database.num_entries() << std::endl;   
      }
      break;   
    }  
    case output_format_arg_xml: {
      std::cout << "<?xml version=\"1.0\"?>" << std::endl << "<maxlink_query2_result xmlns=\"" << MAXLINK_QUERY2_RESULT_NAMESPACE << "\" n1=\"" << n1 << "\" n2=\"" <<  shortest_path_database.num_entries() << "\">" << std::endl;   
      BOOST_FOREACH(maxlink_types::GeneIdHyperGeometricResult it, maxlink_result) {
        // The Coogle C++ Style Guide recommends fprintf instead of iostream     
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=Streams#Streams
        // But if we use printf we also need to take care of 64-bit portability:
        // http://google-styleguide.googlecode.com/svn/trunk/cppguide.xml?showone=64-bit_Portability#64-bit_Portability
        // 
        // Anyway just for now we use iostream
        std::cout << "<gene gene_id=\"" << num2string_gene_database.translateNum2String(it.gene_id) << "\" hypergeometric_probability=\""  << it.hypergeometric_probability << "\" k=\""   << it.k << "\" t=\"" 
                  << it.t << "\"/>" << std::endl;   
      }
      std::cout << "</maxlink_query2_result>" << std::endl;   
      break;   
    }
    default: assert(false); /* indicate programming error */ 
  }
}
